# DTAnalysis (class) #

---

This is essentially the "master" object class for any analysis run using this toolbox. A `DTAnalysis` object contains within it a `DTData` object, a `DTModel` object, and a `DTOutput` object, and it handles a bunch of the basic logic of doing everything within a given analysis.

The `DTAnalysis` object is typically created by `DTJob` from a specification contained in a JSON-format job file, although you can also create one manually in code if you know what you're doing.





<br><br>




## Attributes ##


**nits** *(integer, USUALLY)*: An integer indicating the *number of iterations* of the analysis to run; iterations typically means one complete run of an analysis with a given set of parameters and a particular randomization of training, validation (if necessary), and test sub-datasets. However, the exact meaning may depend on the cross-validation scheme you're using. (See `xval_type`, `nfolds`, `classify_over`, `xval_sa`, etc., below for more details.)


Note that we say `nits` is USUALLY an integer. By this we mean that it should be an integer for all currently fully-implemented cross-validation schemes. For *transfer learning*, a scheme that is currently implemented experimentally but not recommended for anyone outside the dev team to use, `nits` would be a two-item list of integers, with the first item indicating the number of iterations in the "outer" loop and the second indicating the number of iterations in the "inner" loop. (More on this when/if the transfer module goes public -- TK!)


**xval\_type** *(string)*: String code describing the type of cross-validation scheme to use. Currently implemented options (be warned, we did not use the most descriptive names ever) include:

* `single`: Do a single training, validation (if necessary), and test split, train a model, and test it on the test set. This is a pretty common scheme to use if you have a large number of samples (aka trials) and you just want to divvy them up randomly for classification (as opposed to doing classification within subjects or something like that). Good for things like image processing, or for EEG/MEG type analyses where you have a bunch of features (e.g. channels/timepoints) that are essentially the same across subjects, and you want to train/test a "universal" classification model (i.e., one that is trained across trials pulled from all subjects, not trained/tested within subjects). In the *accuracy summary* output of this cross-validation scheme, you will essentially get just one accuracy value per iteration.

* `loop_over_sa`: Do classification separately for each value of a specified "sample attribute." Data samples (aka trials) can be tagged with an arbitrary number of sample attributes, or `sa`'s for short. [pick up here]

* `transfer_over_sa`: EXPERIMENTAL, DO NOT USE unless you're on the dev team or unless Danger is your middle name

* Another option for `xval_type` that is not currently implemented, but will likely be added in the relatively near future, is `kfold` (or something along those lines) for k-fold cross-validation.


**nfolds** *(integer)*: This parameter is actually not currently used but is included in the class for when we get around to implementing k-fold cross-validation. At that time it will indicate how many folds to cross-validate over. Right now it does nothing.


**train\_val\_test** *(3-item list of numeric values)*: This is a list containing the proportions of the input data to use as training, validation, and/or test datasets, in that order. Not all theoretically possible cross-validation schemes require this parameter but all the ones currently implemented do.

The items in this list can be either integers or floating-point values but they have to be numbers and there have to be three of them, even if one of them is not used. (For example, analyses using the PyMVPA backend would not require any validation data, only training/test, but you would still have to specify all three values with the middle value set to 0.) The values should add up to either exactly 100 (i.e., 100 percent) or exactly 1, depending on whether you like percentages or proportions better. If they add up to anything else (within Python's ability to detect, anyway; 99.9 is not good enough but 99.999999999999993 might work), you'll get an error message.

A full discussion of how cross-validation works is beyond the scope of this documentation, but essentially a typical MVPA (e.g., using SVMs or SMLR or something like that) would require you to first generate your model using a "training" dataset and then test it to see how generalizable it is on a held-out "test" dataset. So for a SVM-type analysis, you might specify something like `[80, 0, 20]` for the `train_val_test` parameter, which means 80% of the data will be used for training and then the model will be tested on the held-out 20%. Of course your choices here will affect the quality of your analysis output and depend on lots of factors, but generally speaking something between about a 70/30 split and about a 90/10 split is a safe choice for many datasets.

Note that these proportions may be approximated if your input data do not divide up perfectly according to what numbers you specify, so you don't have to worry about getting things exact. The toolbox will automatically get as close as it can while making sure that (1) you don't end up with a non-integer number of samples (aka trials) in any subset of the data, and (2) all of your classes have an equal number of samples/trials, so as not to bias the classification. This means that a certain number of samples/trials may go unused. (New data subsets will be selected randomly on each iteration of the analysis, so samples/trials that go unused on one iteration will probably get used on a different iteration.)

Keras-based deep learning analyses are similar but they add a "validation" dataset. The validation dataset is like another "test" dataset but used only in training. In other words, in deep learning analyses, the algorithm will typically go through multiple rounds of training the neural network and testing it against the validation dataset before deciding the network is trained up about as well as it's ever going to get. Only then will it perform the "true" test of the network's performance against the held-out test dataset. Again, exactly which proportions work best for you will depend on many factors including how much data you have, but something along the lines of 60/20/20 or 70/20/10 or 80/10/10 is probably what a lot of folks will end up using. (In Python list notation, that would be `[60, 20, 20]` and so on.)


**scaling\_method** *(string or None)*: What type of rescaling operation should be done on your data. Used in conjunction with `scaling` (see below). If both `scaling_method` and `scaling` are `None` (or if you don't specify anything in your JSON configuration file for either of them, as the default for both parameters is `None`), then no data rescaling is performed and analyses are performed on your raw, un-transformed data values. If one of the scaling methods below is specified, then data are rescaled accordingly. Note that rescaling is performed in a way that preserves statistical independence between your training, validation, and/or test datasets, although the particular details of how that rescaling happens depend on which rescaling option you choose. Note also that how much rescaling affects your classification performance is pretty dependent on your analysis architecture and all of your other parameters... sometimes rescaling can make a big difference and other times it doesn't. (In the average case, with data values that aren't too weirdly distributed, it probably makes a small-to-medium difference.) Currently implemented options include:

* `percentile`: All data are rescaled according to the *Nth* percentile of the **absolute value** of all training data points, where N is the value of the `scaling` parameter below.

  This is a little bit weird at first glance but makes sense in many cases. It works best for data that are either non-negative or more-or-less centered around zero (for example, EEG data where zero would mean zero voltage on the scalp relative to the reference electrode). It may not be an ideal choice for datasets that include negative values but whose values are not equally distributed on the positive and negative sides. The idea is that it lets you do something similar to normalizing your values into a pre-defined range like 0-1, but with the normalization based on a moderately-high value in your dataset rather than the very maximum value, which could be a non-representative outlier. And by "high," we mean in absolute value terms (distance from zero) -- for scaling purposes we don't really care if the most extreme values are on the positive or the negative end. Although again, things could get funky if you have both positive and negative values in your data but they aren't at least *approximately* equally distributed around zero.

  As an example, if you use this `scaling_method` value and set `scaling` (see below) to 80, all the values in the dataset would be divided by whatever value represents the 80th percentile of the magnitudes (absolute values) of the values in your training dataset. Note that we do use the value calculated from the percentiles of the training dataset to scale the validation and test datasets as well, because the training dataset is often the largest and thus most statistically robust. However, as all we're doing is dividing all data by a common scale factor (and the characteristics of the validation/test datasets are not used in the percentile calculations), this should not cause a problem for the statistical independence of the training, validation, and test datasets. The reason we do this is because we don't want statistical fluctuations to cause different scale factors to be applied to the training, validation, and test datasets.

  Note that, for boring historical reasons (basically, so some very old scripts won't break), this is the default rescaling method chosen if you specify `None` for `scaling_method` but **DO** put a value in for `scaling`. We don't exactly suggest you do that though, because it's probably better to explicitly specify `scaling_method` for the sake of clarity even if you do intend to use the `percentile` option. Just giving you the heads-up.

  Also, to reiterate -- the percentile calculation is done over all training data points, not separately for each feature (voxel or electrode/timepoint or whatever). So the assumption is that all of your data points have SOMETHING in common that makes it reasonable to scale the data across all features. If you want to rescale in a more fine-grained manner, you might have to do it yourself before reading the data into this toolbox -- currently we have no implemented way to scale the data WITHIN features, just because there are a million ways you might want to do it, many of which could require knowing something about the structure of the dataset that would break the toolbox's minimal set of assumptions about what you're going to hand it. This is really just a basic rescaling to get values into a semi-standardized range... if you want anything fancier you might have to roll your own (or request it as a feature for a future toolbox release).

* `standardize`: Similar to `percentile` above but Z-score instead of doing a percentile-based scaling. Most of what we said above for `percentile` applies to `standardize` as well -- for example, all scaling is done based on values from the training dataset, and scaling is performed across all data points (not separately for each feature). So, see `percentile` above for all the quirks and caveats... this one is pretty similar in everything but the actual way we rescale the data.

  Note: the `scaling` value (see below) is not used for this scaling method; if you specify anything, it is silently ignored.

* `map_range`: Scales all training data into a given range, which by default is zero to one. In other words, for the default 0-1 range, this subtracts the minimum value in the training dataset from all training dataset values, then divides all values by the largest value in the post-subtraction dataset, to get everything into the 0-1 range. The idea is similar for other ranges, but you can also scale your data from 0 to 100 or -40 to 137 or whatever you please. To specify a minimum/maximum other than 0 and 1, make the `scaling` value (see below) a two-item list where both items are numeric values. The first value will be taken as the minimum of the mapped range, and the second value will specify the maximum.

  Note that, as with `percentile` and `standardize`, the mapping is done based on the values in the training dataset. So your validation and/or test datasets may end up with values outside the 0-1 range (or whatever non-default range you specify), or they may not contain the ends of the range. But the scaling factors that are applied to validation and test datasets will be the same scaling parameters that are applied to the training dataset.

* `mean_center`: Like Z-scoring but no dividing by the standard deviation -- just subtracting out the mean. As with other methods, this is done across ALL data points in the training dataset, and the value calculated from the training dataset will also be applied to the validation and test datasets. (So, you will be guaranteed that the values in the training dataset will end up centered around 0, but the validation and/or test datasets may not be exactly centered around 0 themselves. Although if the distributions of your validation and test datasets are pretty similar to the distribution of your training dataset, things should come *pretty* close to being centered around 0.) The `scaling` value (see below) is ignored for this too.


**scaling** *(single numeric value, list of 2 numeric values, or None)*: Specifies any numeric information needed to implement any of the `scaling_method` options listed above. For `percentile` it's a single number indicating what percentile, for `standardize` this value is ignored, for `map_range` it's a two-item list specifying the minimum and maximum values in the range, and for `mean_center` this value is ignored. If the value is unspecified or specified as `None`, semi-reasonable defaults chosen by your benevolent development overlords will be used.


**classify\_over** *(string, probably)*: The name of whatever "sample attribute," or SA for short (see `DTData` documentation for more info), you want to do your classification over. In theory this doesn't HAVE to be a string -- we don't do a lot of checking so I guess it could be anything that Python allows you to use as a key to a dictionary -- but if you are vaguely sane it will almost always be a string.

To briefly summarize the `DTData` stuff, each sample (aka trial) in your dataset can be tagged with one or more SA's. These could be things like class, or subject, or run, or really any attribute that a sample/trial could have. Each of these SAs has a name that you give it when you load in your data.

Anyway, `classify_over` is just the name of whatever SA you want to run your classification on. For example, say I have some data where people are looking at faces or scenes in either the left or right side of their visual field. I might have those attributes tagged with SAs named `category` and `field` (you can use whatever names you want). If I want to run a classification of faces vs scenes, I would specify `category` for `classify_over`. If I want to classify left vs right instead, I could just change `classify_over` to `field`. Easy, right?!


**backend\_options** *(dictionary or None)*: A set of options that are essentially passed straight along to whatever backend you're using, in theory. In practice, at the moment only the Keras backend has any options that use this parameter -- if you are using the PyMVPA backend then `backend_options` just gets ignored. And we have kind of renamed a bunch of them because they're used in different places within Keras that we're trying to abstract away from you, dear user, by processing them ourselves at appropriate junctures, so don't go looking for things named exactly like these options in the Keras source code.

We think these more-or-less global options make more sense as part of `DTAnalysis` than any other class, so we typically specify them here -- although our dirty little secret is that they basically just get passed along to `DTModel` for most of the heavy lifting. So you might find more info in the `DTModel` documentation. (And if you're curious about how we parse these options and send them to Keras under the hood, look at `DTModel.py` for the down-and-dirty.)

But in brief, currently implemented options include the following:

* `monitor` *(string)*: What to monitor for Keras's early stopping (`keras.callbacks.EarlyStopping`). Our default is `val_loss`. TK more details on all of these.

* `patience` *(integer)*: Patience parameter for Keras's early stopping (ditto). Our default is `50`.

* `checkpoint_filename` *(string)*: The name of a temporary file that we have to write out during Keras analyses. Not that critical for you to specify unless you truly feel passionately about your tempfile names. We should generally clean up after ourselves but in case one of these does not get cleared away for some reason, you may want to know that the default is `delineate_checkpoint_weights_tempfile` plus a "hash" string calculated from the job file you're running. So if you see such a file lying around it is safe to delete (assuming you don't have an analysis running currently... if you do have an analysis running then leave it be, and it will probably get cleaned up on its own later if the analysis finishes normally).

* `batch_size` *(integer)*: The size of one "batch" in Keras. (TK: define better.) Our default is `500` but you may want/need to tweak this based on GPU available RAM and/or other factors.

* `epochs` *(integer)*: How many times to train your model. We default to `10000` but in practice it may not matter because we always have early-stopping on (i.e. stop training the neural network when it seems like learning has plateaued) and we almost never get to this number of epochs before early-stopping kicks in. This parameter may become more relevant to users if we ever implement the option to disable early stopping.

* `verbose` *(integer)*: Keras verbosity level. Default is `1`. (TK: more detail.) Note that we currently have a Bitbucket issue on the list to replace this with a more toolbox-general verbosity setting so don't get too attached.

* `reduction_factor` *(numeric value)*: How much to reduce the learning rate by each time learning rate reduction kicks in (TK more detail). (Keras 2+ only, I believe.)

* `reduction_patience`: This is currently an "Easter egg" by which I mean it doesn't appear to do anything. Oops. Hopefully we'll fix this in the next commit or two.


Again, these options are presently for **KERAS ONLY** -- they will be silently ignored in PyMVPA. Formatting them in a JSON job file might be a bit confusing at first (dictionary syntax is hard), so you might want to check out the `sample_jobfiles` directory for some usage hints.




backend_options=None, 

dataset=None, 

model=None, 

xval_sa = None, 

output_handler = None




<br><br>


## Methods ##




<br><br>


## JSON job file options ##






<br><br>



Just grabbing docs and making notes for now, will convert to true Markdown later.


    """
    
    The master object

    Proposed attributes (from outline on whiteboard that we took a photo of and posted to Slack):
    
    dataset (type: DTData) -- a DTData object
    model (type: DTModel) -- actually, could be a DTModel or an object subclassed from DTModel (e.g. we may want to have different subtypes of models)
    what_to_save (type: string, or list of strings?) -- one or more signifiers of what output we want to save... presumably this would be things like 
        'accuracies' and/or 'aucs' (area under ROC curves) and/or 'confusion_matrices' and/or 'all_outputs' (e.g. all the scores produced by the 
        classifier), and/or so on. This attribute might end splitting into more than one attribute, or maybe something like a dictionary, potentially?
        E.g. a series of Boolean flags, or certain outputs might require additional arguments -- for example, what filetype to use when saving output
        (CSV vs tab-delimited text aka TSV vs JSON vs whatever).
    output_file (type: string) -- what to call the output file. Will have to see how this goes in implementation -- it might be better to just make 
        it the first part of an output filename (e.g. if there are multiple outputs, you might have 'my_output_accuracies.csv' and 'my_output_aucs.csv'
        and so on). Or we could switch this over to being the name of a directory to save the output in, if we decide we're pretty much always going 
        to have multiple output files. Or we could make file-vs-directory yet another user-configurable option.
    nits (type: integer) -- number of iterations to run, i.e. number of times to train/test the model, throw it out, initialize a new model, re-divvy 
        up the data into different training/validation/test sets, and run again. Some decisions will have to be made as to how to treat this attribute 
        for different analysis types -- for example, if we are doing classification on a per-subject basis, do we do all the iterations for each 
        subject first and then do the same number of iterations for the next subject, or do we do one iteration for each subject and then do a second 
        pass through all the subjects, etc.? Or make this configurable as well?
    type (type: string) -- what overall class of analysis to run. Some basic examples could be 'universal' (across all subjects), 'subject_wise' 
        (train and test a separate model for each subject), 'transfer' (train on a subset of the data and then lop off part of the classifier and 
        re-train on the remaining data), or maybe others. Some of these (e.g. 'transfer') might have to be something like a dictionary (with more 
        options specified inside) rather than just a simple string. Or we could always have the option to pass in something like a function handle 
        if the user has something more exotic in mind, and that function would figure out what to do. (The same goes for a lot of the options above, 
        actually... a good fallback plan for many of them might be to allow a handle to a function with a given specification to be passed in, so 
        the user can always overload most of the default functionalities pretty easily while still using our basic framework.)
    t_v_t (type: tuple of integers/floats) -- percentages of data to use as training, validation, and test datasets, respectively. Actually, no need 
        for these to be percentages per se... they could just be relative values expressing the ratios of those sets. So, for example, (80, 10, 10) 
        and (8, 1, 1) and (4, .5, .5) would all represent the same thing, 80% training, 10% validation, 10% test. No need to make the user do the 
        percent math if they want to think ratio-wise. Presumably, the validation and test numbers could be zero, either if we don't want to hold out
        a test dataset for some reason, or if validation is not a concept that makes sense for the classifier in question. (E.g. an SVM would probably
        just have train/test.) This attribute might be ignored for certain values of 'xval_type' below because it doesn't really make sense with all 
        of them.
    xval_type (type: string) -- cross-validation scheme. For example, 'kfold_rand' to do k-fold cross-validation with random selection of folds (this 
        presumably makes more sense for simple classifiers that just need train/test data... would have to figure out how to interpret this with 
        validation in the mix... also we might need an additional parameter or to make this a dictionary or something in order to indicate the # of 
        folds), 'single' to just do a single (randomized) divvying-up of training/validation/test in each iteration, 'kfold_session' to train/test (and 
        maybe validate) across some kind of session-like natural division in the data (e.g. runs in fMRI, or could be something like subjects... we may 
        need an additional parameter or to make this attribute a dictionary so we can specify exactly which labels in the data we want to treat as the 
        'session' variable).
    
    Proposed methods:
    
    run() -- the most important one. Runs the analysis and does pretty much everything. We haven't gotten much beyond this but presumably implementing 
        it will entail creating a bunch of other helper methods, some of which will make sense to make user-accessible. For example, one we wrote down 
        was check_already_ran(), which would presumably look for the specified output and see if it appeared that the analysis in question had already 
        happened. (This could be useful in implementing a batch system -- e.g. if a batch runs through a bunch of analyses but then crashes or gets 
        interrupted somehow, you don't want to have to manually check where to resume... you'd rather just start the batch running again and let it 
        figure out where it left off.)
    validate_arguments() -- all major classes will need one of these to help DTBatch decide if batch file structures are valid. Takes in a dict of 
        keys/values, which are presumed to represent input arguments to this class's initializer. Returns True if valid and False if not.
    
    
    """